import os
from typing import (
    ClassVar,
    List,
    Optional,
    Type,
    TYPE_CHECKING,
)

from cryptography.x509 import load_pem_x509_certificate
from cryptography.hazmat.primitives.serialization import load_pem_private_key

from .base import (
    Check,
    CheckInvoker,
    MissingPropertyCheck,
    PropertyConstraintCheck,
)

if TYPE_CHECKING:
    from ..models.globus_oidc_idp import GlobusOIDCIdP  # noqa: F401


class GlobusOIDCIdPCheck(CheckInvoker):
    checkers: ClassVar[List[Type[Check["GlobusOIDCIdP"]]]] = []
    object_name = "globus_oidc_idp"


@GlobusOIDCIdPCheck.property_checker(property_name="display_name")
class GO01(MissingPropertyCheck["GlobusOIDCIdP"]):
    ...


@GlobusOIDCIdPCheck.property_checker(property_name="display_name")
class GO02(PropertyConstraintCheck["GlobusOIDCIdP"]):
    constraint_name = "value contains all whitespace"

    def value_is_ok(self, value: Optional["GlobusOIDCIdP"]) -> bool:
        ok = True
        if value and value.display_name is not None:
            display_name = value.display_name.strip()
            if not display_name:
                ok = False
        return ok


@GlobusOIDCIdPCheck.property_checker(property_name="support_email")
class GO03(MissingPropertyCheck["GlobusOIDCIdP"]):
    ...


@GlobusOIDCIdPCheck.property_checker(property_name="support_email")
class GO04(PropertyConstraintCheck["GlobusOIDCIdP"]):
    constraint_name = "must be set"

    def value_is_ok(self, value: Optional["GlobusOIDCIdP"]) -> bool:
        ok = True
        if value and value.support_email is not None:
            support_email = value.support_email.strip()
            if not support_email:
                ok = False
        return ok


@GlobusOIDCIdPCheck.property_checker(property_name="support_contact")
class GO05(MissingPropertyCheck["GlobusOIDCIdP"]):
    ...


@GlobusOIDCIdPCheck.property_checker(property_name="support_contact")
class GO06(PropertyConstraintCheck["GlobusOIDCIdP"]):
    constraint_name = "must be set"

    def value_is_ok(self, value: Optional["GlobusOIDCIdP"]) -> bool:
        ok = True
        if value and value.support_contact is not None:
            support_contact = value.support_contact.strip()
            if not support_contact:
                ok = False
        return ok


@GlobusOIDCIdPCheck.property_checker(property_name="server_name")
class GO07(MissingPropertyCheck["GlobusOIDCIdP"]):
    ...


@GlobusOIDCIdPCheck.property_checker(property_name="server_name")
class GO08(PropertyConstraintCheck["GlobusOIDCIdP"]):
    constraint_name = "must be set"

    def value_is_ok(self, value: Optional["GlobusOIDCIdP"]) -> bool:
        ok = True
        if value and value.server_name is not None:
            server_name = value.server_name.strip()
            if not server_name:
                ok = False
        return ok


@GlobusOIDCIdPCheck.property_checker(property_name="certificate_path")
class GO09(MissingPropertyCheck["GlobusOIDCIdP"]):
    ...


@GlobusOIDCIdPCheck.property_checker(property_name="certificate_path")
class GO10(PropertyConstraintCheck["GlobusOIDCIdP"]):
    constraint_name = "must be set"

    def value_is_ok(self, value: Optional["GlobusOIDCIdP"]) -> bool:
        ok = True
        if value and value.certificate_path is not None:
            certificate_path = value.certificate_path.strip()
            if not certificate_path:
                ok = False
        return ok


@GlobusOIDCIdPCheck.property_checker(property_name="certificate_chain_path")
class GO11(MissingPropertyCheck["GlobusOIDCIdP"]):
    ...


@GlobusOIDCIdPCheck.property_checker(property_name="certificate_chain_path")
class GO12(PropertyConstraintCheck["GlobusOIDCIdP"]):
    constraint_name = "must be set"

    def value_is_ok(self, value: Optional["GlobusOIDCIdP"]) -> bool:
        ok = True
        if value and value.certificate_chain_path is not None:
            certificate_chain_path = value.certificate_chain_path.strip()
            if not certificate_chain_path:
                ok = False
        return ok


@GlobusOIDCIdPCheck.property_checker(property_name="private_key_path")
class GO13(MissingPropertyCheck["GlobusOIDCIdP"]):
    ...


@GlobusOIDCIdPCheck.property_checker(property_name="private_key_path")
class GO14(PropertyConstraintCheck["GlobusOIDCIdP"]):
    constraint_name = "must be set"

    def value_is_ok(self, value: Optional["GlobusOIDCIdP"]) -> bool:
        ok = True
        if value and value.private_key_path is not None:
            private_key_path = value.private_key_path.strip()
            if not private_key_path:
                ok = False
        return ok


@GlobusOIDCIdPCheck.property_checker(property_name="certificate_path")
class GO15(PropertyConstraintCheck["GlobusOIDCIdP"]):
    constraint_name = "certificate file must exist"

    def value_is_ok(self, value: Optional["GlobusOIDCIdP"]) -> bool:
        ok = True
        if value and value.certificate_path:
            certificate_path = value.certificate_path
            if not os.path.exists(certificate_path):
                ok = False

        return ok


@GlobusOIDCIdPCheck.property_checker(property_name="certificate_chain_path")
class GO16(PropertyConstraintCheck["GlobusOIDCIdP"]):
    constraint_name = "certificate chain file must exist"

    def value_is_ok(self, value: Optional["GlobusOIDCIdP"]) -> bool:
        ok = True
        if value and value.certificate_chain_path:
            certificate_chain_path = value.certificate_chain_path
            if not os.path.exists(certificate_chain_path):
                ok = False

        return ok


@GlobusOIDCIdPCheck.property_checker(property_name="private_key_path")
class GO17(PropertyConstraintCheck["GlobusOIDCIdP"]):
    constraint_name = "private key file must exist"

    def value_is_ok(self, value: Optional["GlobusOIDCIdP"]) -> bool:
        ok = True
        if value and value.private_key_path:
            private_key_path = value.private_key_path
            if not os.path.exists(private_key_path):
                ok = False

        return ok


@GlobusOIDCIdPCheck.property_checker(property_name="certificate_path")
class GO18(PropertyConstraintCheck["GlobusOIDCIdP"]):
    constraint_name = "certificate path contains a PEM certificate"

    def value_is_ok(self, value: Optional["GlobusOIDCIdP"]) -> bool:
        ok = True
        if value and value.manage_certificate_and_key and value.certificate_path:
            certificate_path = value.certificate_path
            if os.path.exists(certificate_path):
                try:
                    with open(certificate_path, "rb") as f:
                        data = f.read()
                    load_pem_x509_certificate(data)
                except IOError:
                    ok = False
                except ValueError:
                    ok = False

        return ok


@GlobusOIDCIdPCheck.property_checker(property_name="certificate_path")
class GO19(PropertyConstraintCheck["GlobusOIDCIdP"]):
    constraint_name = "key path contains a PEM private key"

    def value_is_ok(self, value: Optional["GlobusOIDCIdP"]) -> bool:
        ok = True
        if value and value.manage_certificate_and_key and value.private_key_path:
            private_key_path = value.private_key_path
            if os.path.exists(private_key_path):
                try:
                    with open(private_key_path, "rb") as f:
                        data = f.read()
                    load_pem_private_key(data, password=None)
                except IOError:
                    ok = False
                except ValueError:
                    ok = False

        return ok


@GlobusOIDCIdPCheck.property_checker(property_name="certificate_chain_path")
class GO20(PropertyConstraintCheck["GlobusOIDCIdP"]):
    constraint_name = "certificate chain path contains a PEM certificate"

    def value_is_ok(self, value: Optional["GlobusOIDCIdP"]) -> bool:
        ok = True
        if value and value.manage_certificate_and_key and value.certificate_chain_path:
            certificate_chain_path = value.certificate_chain_path
            if os.path.exists(certificate_chain_path):
                try:
                    with open(certificate_chain_path, "rb") as f:
                        data = f.read()
                    load_pem_x509_certificate(data)
                except IOError:
                    ok = False
                except ValueError:
                    ok = False

        return ok


# TODO: verify that certificate matches server_name
