from typing import (
    ClassVar,
    List,
    Optional,
    Set,
    TYPE_CHECKING,
    Type,
)

from ..util import get_api

from .base import (
    Check,
    CheckInvoker,
    ObjectIsNotNoneCheck,
    MissingPropertyCheck,
    PropertyConstraintCheck,
)

if TYPE_CHECKING:
    from ..models import HostEndpoint  # noqa: F401


class HostEndpointCheck(CheckInvoker["HostEndpoint"]):
    checkers: ClassVar[List[Type[Check["HostEndpoint"]]]] = []
    object_name = "endpoint"


@HostEndpointCheck.checker
class EP01(ObjectIsNotNoneCheck["HostEndpoint"]):
    ...


@HostEndpointCheck.property_checker(property_name="organization")
class EP02(MissingPropertyCheck["HostEndpoint"]):
    ...


@HostEndpointCheck.property_checker(property_name="owner")
class EP03(MissingPropertyCheck["HostEndpoint"]):
    ...


@HostEndpointCheck.property_checker(property_name="owner")
class EP04(PropertyConstraintCheck["HostEndpoint"]):
    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super().__init__(*args, **kwargs)  # type: ignore
        self.missing_identities: Set[str] = set()

    def value_is_ok(self, value: Optional["HostEndpoint"]) -> bool:
        ok = True
        if value and value.owner is not None:
            owner = value.owner
            api = get_api("auth")
            res = api.get_identities(ids=owner)

            identities = res["identities"]

            if len(identities) != 1:
                self.missing_identities.add(owner)
                ok = False
        return ok

    constraint_name = "identity doesn't exist in Auth"


@HostEndpointCheck.property_checker(property_name="owner")
class EP05(PropertyConstraintCheck["HostEndpoint"]):
    def value_is_ok(self, value: Optional["HostEndpoint"]) -> bool:
        ok = True
        if value:
            owner = value.owner
            api = get_api("auth")
            res = api.get_identities(ids=owner)

            identities = res["identities"]
            if (
                identities
                and identities[0]
                and identities[0]["identity_type"] == "email"
            ):
                ok = False

        return ok

    constraint_name = "identity uses the email provider"
