from abc import abstractmethod
from typing import (
    ClassVar,
    Generic,
    List,
    Optional,
    Type,
    TypeVar,
    TYPE_CHECKING,
    Union,
)

from .base import (
    Check,
    CheckInvoker,
    MissingPropertyCheck,
    ObjectIsNotNoneCheck,
)

if TYPE_CHECKING:
    from ..models import (  # noqa: F401
        AuthIdP,
        Credential,
        GlobusOIDCIdP,
        MigrationPlan,
        StorageGateway,
    )

T = TypeVar("T")


class RelationshipChecker(Generic[T], Check["MigrationPlan"]):
    """
    Implementation of Check that checks relationships between multiple
    objects, so that either could be fixed to eliminate the violation
    """

    relationship_object_names: ClassVar[List[str]]
    constraint_name: ClassVar[str]

    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        self.violations: List[List[T]] = []

    def is_ok(self, obj: Optional["MigrationPlan"]) -> bool:
        ok = True
        err_objs: List[List[T]] = []
        if obj is not None:
            err_objs = self.relationship_is_ok(obj)
        if err_objs:
            for err_obj in err_objs:
                if err_obj not in self.violations:
                    self.violations.append(err_obj)
            ok = False
        return ok

    @abstractmethod
    def relationship_is_ok(self, obj: "MigrationPlan") -> List[List[T]]:
        ...

    @property
    def relationship_object_name_list(self) -> str:
        if len(self.relationship_object_names) < 2:
            return str(self.relationship_object_names)
        elif len(self.relationship_object_names) == 2:
            return " and ".join(self.relationship_object_names)
        else:
            return (
                ", ".join(self.relationship_object_names[:-1])
                + ", and "
                + self.relationship_object_names[-1]
            )

    @property
    def diagnostic(self) -> Optional[str]:
        if self.violations:
            violation_strings: List[str] = []
            for v in self.violations:
                violation_str = ", ".join([str(w) for w in v])
                violation_strings.append(f"({violation_str})")
            return (
                f'The relationship "{self.constraint_name}" between '
                f"{self.relationship_object_name_list} is invalid for "
                f"{violation_strings}"
            )
        return None


class MigrationPlanCheck(CheckInvoker["MigrationPlan"]):
    checkers: ClassVar[List[Type[Check["MigrationPlan"]]]] = []
    object_name = "migration_plan"


@MigrationPlanCheck.checker
class MP01(ObjectIsNotNoneCheck["MigrationPlan"]):
    ...


@MigrationPlanCheck.property_checker(property_name="storage_gateway")
class MP02(MissingPropertyCheck["MigrationPlan"]):
    ...


@MigrationPlanCheck.property_checker(property_name="mapped_collection")
class MP03(MissingPropertyCheck["MigrationPlan"]):
    ...


@MigrationPlanCheck.checker
class MP04(
    RelationshipChecker[
        Union["StorageGateway", str, List["AuthIdP"], Optional["GlobusOIDCIdP"]]
    ]
):
    relationship_object_names = [
        "storage_gateway",
        "domain",
        "auth_idp",
        "globus_oidc_idp",
    ]
    constraint_name = "allowed_domains must be a from an existing Auth IdP or from a Globus OIDC server"

    def relationship_is_ok(
        self, obj: "MigrationPlan"
    ) -> List[
        List[Union["StorageGateway", str, List["AuthIdP"], Optional["GlobusOIDCIdP"]]]
    ]:
        violations: List[
            List[
                Union["StorageGateway", str, List["AuthIdP"], Optional["GlobusOIDCIdP"]]
            ]
        ] = []

        sg = obj.storage_gateway
        if not sg:
            return violations

        if not sg.allowed_domains:
            return violations

        if obj.identity_providers:
            identity_providers = list(obj.identity_providers.values())
        else:
            identity_providers = []
        oidc_server = obj.oidc_server

        if not identity_providers and not oidc_server:
            return violations

        for domain in sg.allowed_domains:
            if identity_providers:
                found_in_idp = False
                for idp in identity_providers:
                    if idp.domains and domain in idp.domains:
                        found_in_idp = True
                        break
                if found_in_idp:
                    continue
            if oidc_server:
                if domain == oidc_server.server_name:
                    continue
            violations.append([sg, domain, identity_providers, oidc_server])

        return violations


@MigrationPlanCheck.checker
class MP05(RelationshipChecker[Union["Credential", "StorageGateway"]]):
    relationship_object_names = ["user_credential", "storage_gateway"]
    constraint_name = "identities must be a domain allowed by the storage gateway"

    def relationship_is_ok(
        self, obj: "MigrationPlan"
    ) -> List[List[Union["Credential", "StorageGateway"]]]:
        violations: List[List[Union["Credential", "StorageGateway"]]] = []

        sg = obj.storage_gateway
        if not sg:
            return violations
        if not sg.allowed_domains:
            return violations
        gc_list = obj.guest_collections
        if not gc_list:
            return violations

        for gc in gc_list:
            uc = gc.credential
            if uc:
                if uc.identity_username:
                    if uc.identity_username.rsplit("@", 1)[1] not in sg.allowed_domains:
                        violations.append([uc, sg])

        return violations
