from typing import Optional
import uuid

import click

from ..util import get_api
from ..models import AuthIdP, MigrationPlan


@click.command("add-idp")
@click.option("--id", type=click.UUID, help="ID of the identity provider")
@click.option("--domain", help="Domain of the identity provider")
@click.option(
    "--all-domains",
    is_flag=True,
    help="When using the --domain option, this will add mappings for the IdP for "
    "all domains provided by this idp, instead of just the named domain",
)
def add_idp(id: Optional[uuid.UUID], domain: Optional[str], all_domains: bool) -> None:
    migration_plan = MigrationPlan.load()

    if not (id or domain):
        raise click.ClickException("Missing required parameter --id or --domain")

    auth_api = get_api(service="auth")

    if id:
        existing_idp = migration_plan.get_identity_provider(id=str(id))
        if existing_idp:
            click.echo(
                f"IdP {existing_idp.id} already exists and is configured for domains "
                f"{existing_idp.domains}"
            )
        else:
            res = auth_api.get("/v2/api/identity_providers/{id}")
            idp = res.get("identity_provider")
            if idp:
                auth_idp = AuthIdP.from_identity_provider_document(idp)
                migration_plan.add_identity_provider(auth_idp)
    elif domain:
        if not all_domains:
            existing_idp = migration_plan.get_identity_provider_for_domain(domain)
            if existing_idp:
                click.echo(
                    f"IdP {existing_idp.id} already exists and is configured for "
                    f"domains {existing_idp.domains}"
                )
        if not existing_idp:
            res = auth_api.get(
                "/v2/api/identity_providers", params={"domains": [domain]}
            )
            for idp in res["identity_providers"]:
                idp_id = idp["id"]
                if not all_domains:
                    idp["domains"] = [d for d in idp["domains"] if d == domain]
                if idp["domains"]:
                    existing_idp = migration_plan.get_identity_provider(id=idp_id)
                    if existing_idp:
                        existing_idp.domains = list(
                            set(existing_idp.domains or {}).union(idp["domains"])
                        )
                        existing_idp.mapping = None
                    else:
                        auth_idp = AuthIdP.from_identity_provider_document(idp)
                        migration_plan.add_identity_provider(auth_idp)
    migration_plan.dump_identity_providers()


if __name__ == "__main__":
    add_idp()
