from typing import Dict, List, Set, Tuple

import click

from ..models import AuthIdP, MigrationPlan
from ..util import get_api


@click.command("add-idp-for-cilogon")
@click.option(
    "--domain",
    metavar="DOMAIN",
    multiple=True,
    help="Domain to use from this IdP. Use this if the configured "
    "organization(s) supports multiple domains and you only want to use a subset "
    "of them. This option can be passed multiple times to include multiple "
    "domains. If not passed and there are multiple domains, then the idp will be "
    "configured for all domains.",
)
def add_idp_for_cilogon(domain: Tuple[str]) -> None:
    migration_plan = MigrationPlan.load()

    domains: Set[str] = set()

    if migration_plan.gsi_authz_configuration:
        for callout in migration_plan.gsi_authz_configuration.cilogon_callouts:
            if callout.domains:
                domains.union(callout.domains)

    # If one or more --domain values are passed in, only use those domains
    if domain:
        domains = domains.intersection(domain)

    auth_api = get_api(service="auth")

    pending_domains = set(domains)
    for d in pending_domains:
        existing_idp = migration_plan.get_identity_provider_for_domain(d)
        if existing_idp:
            continue

        res = auth_api.get(
            "/v2/api/identity_providers", params={"domains": list(domains)}
        )
        for idp in res["identity_providers"]:
            existing_idp = migration_plan.get_identity_provider(id=idp["id"])
            if existing_idp:
                existing_idp.add_domain(d)
            else:
                idp_domains = pending_domains.intersection(idp["domains"])

                if idp_domains:
                    idp["domains"] = list(idp_domains)
                    auth_idp = AuthIdP.from_identity_provider_document(idp)
                    migration_plan.add_identity_provider(auth_idp)

    dn_to_domains: Dict[str, List[str]] = {}
    if migration_plan.gsi_authz_configuration:
        for callout in migration_plan.gsi_authz_configuration.cilogon_callouts:
            if callout.domains:
                callout_domains = domains.intersection(callout.domains)
            else:
                callout_domains = None
            if callout_domains:
                if callout.dn not in dn_to_domains:
                    dn_to_domains[callout.dn] = []
                for cd in callout_domains:
                    if cd not in dn_to_domains[callout.dn]:
                        dn_to_domains[callout.dn].append(cd)

    migration_plan.dump_identity_providers()


if __name__ == "__main__":
    add_idp_for_cilogon()
