import enum
from typing import (
    cast,
    Any,
    Generator,
    Iterable,
    List,
)

import click

import gcs_migration.check as checks
import gcs_migration.models as models
from gcs_migration.models import MigrationPlan


class ObjectTypeChoices(enum.Enum):
    acl = "acl"
    credential = "credential"
    endpoint = "endpoint"
    guest_collection = "guest-collection"
    mapped_collection = "mapped-collection"
    migration_plan = "migration-plan"
    oidc_server = "oidc-server"
    role = "role"
    storage_gateway = "storage-gateway"


def get_acls(m: MigrationPlan) -> Generator[models.ACL, None, None]:
    if m.guest_collections:
        for g in m.guest_collections:
            if g.acls:
                for a in g.acls:
                    yield a


def get_credentials(m: MigrationPlan,) -> Generator[models.Credential, None, None]:
    if m.guest_collections:
        for g in m.guest_collections:
            if g.credential:
                yield g.credential


def get_guest_collections(
    m: MigrationPlan,
) -> Generator[models.GuestCollection, None, None]:
    if m.guest_collections:
        for g in m.guest_collections:
            yield g


def get_roles(m: MigrationPlan) -> Generator[models.Role, None, None]:
    if m.host_endpoint and m.host_endpoint.roles:
        for role in m.host_endpoint.roles:
            yield role
    if m.guest_collections:
        for g in m.guest_collections:
            if g.roles:
                for role in g.roles:
                    yield role


@click.command(
    "check",
    short_help="Check migration plan for consistency",
    help="""\
Check the migration plan for consistency, verifying that all identity mappings
are covered by the configuration, and that all configuration settings are
covered by this tool.

If not given any options, all objects are checked. Otherwise, only the objects
of the specified types are checked.
""",
)
@click.option(
    "--type-check",
    multiple=True,
    type=click.Choice([e.value for e in ObjectTypeChoices], case_sensitive=False),
)
def check(type_check: Iterable[str]) -> None:
    migration_plan = MigrationPlan.load()

    object_checkers = {
        ObjectTypeChoices.acl: (checks.ACLCheck, get_acls),
        ObjectTypeChoices.credential: (checks.CredentialCheck, get_credentials),
        ObjectTypeChoices.oidc_server: (
            checks.GlobusOIDCIdPCheck,
            lambda x: [x.oidc_server],
        ),
        ObjectTypeChoices.guest_collection: (
            checks.GuestCollectionCheck,
            get_guest_collections,
        ),
        ObjectTypeChoices.endpoint: (
            checks.HostEndpointCheck,
            lambda x: [x.host_endpoint],
        ),
        ObjectTypeChoices.mapped_collection: (
            checks.MappedCollectionCheck,
            lambda x: [x.mapped_collection],
        ),
        ObjectTypeChoices.migration_plan: (checks.MigrationPlanCheck, lambda x: [x]),
        ObjectTypeChoices.role: (checks.RoleCheck, get_roles),
        ObjectTypeChoices.storage_gateway: (
            checks.StorageGatewayCheck,
            lambda x: [x.storage_gateway],
        ),
    }

    rc = 0

    if not type_check:
        type_check = [e.value for e in ObjectTypeChoices]

    for tc in type_check:
        check_name = ObjectTypeChoices[tc.replace("-", "_")]
        click.echo(f"Checking {check_name.name}...")
        checker_class, fetcher = object_checkers[check_name]

        checker = checker_class()
        violations: List[checks.Check[Any]] = []

        for datum in cast(Iterable, fetcher(migration_plan)):
            res = checker.check(datum)
            if res:
                rc += len(res)
                for r in res:
                    if r not in violations:
                        violations.append(r)
        for v in violations:
            print(f"{v.name}: {v.diagnostic}")

    click.get_current_context().exit(rc)
