import pathlib
import time
from typing import Any, cast, Dict, Iterable, Optional

import click
import globus_sdk

from ..models import (
    Credential,
    CredentialDict,
    GuestCollection,
    HostEndpoint,
    MappedCollection,
    MigrationPlan,
    StorageGateway,
)
from ..parsers import GCSConfig, GridFTPOptionParser
from ..util import get_api, AuthorizeSharing
from ..version import __version__


class RateLimitedTransferClient(globus_sdk.TransferClient):
    __pause_limit: int
    __api_calls: int

    def __new__(
        cls, api: globus_sdk.TransferClient, pause_limit: int = 10
    ) -> "RateLimitedTransferClient":
        obj = api

        obj.__pause_limit = pause_limit
        obj.__api_calls = 0

        return cast(RateLimitedTransferClient, obj)

    def get(self, *args, **kwargs) -> Any:  # type: ignore
        if self.__api_calls > self.__pause_limit:
            time.sleep(1)
            self.__api_calls = 0
        return super().get(*args, **kwargs)


@click.command("create")
@click.version_option(version=__version__)
def create() -> None:
    gcs_config = GCSConfig()
    epid = gcs_config.endpoint_id

    while not epid:
        endpoint_uuid = click.prompt("Enter endpoint ID", type=click.UUID)
        epid = str(endpoint_uuid)

    click.echo("Parsing endpoint state...")
    try:
        gridftp_config = GridFTPOptionParser(
            config_file="/etc/gridftp.conf", config_dir="/etc/gridftp.d"
        )
    except FileNotFoundError as e:
        click.echo(f"WARNING: Unable to load gridftp configuration file: {e}", err=True)
        gridftp_config = GridFTPOptionParser()

    migration_plan = MigrationPlan(".")
    try:
        migration_plan.gcs_config = gcs_config
        migration_plan.gridftp_config = gridftp_config
    except Exception as e:
        click.echo(
            f"Unable to parse the local configuration files: {e}.\nmigration plan data will be incomplete...",
            err=True,
        )

    api = RateLimitedTransferClient(get_api())

    api = RateLimitedTransferClient(get_api())

    res = api.get(f"/endpoint/{epid}")

    transfer_endpoint = res.data

    host_endpoint = HostEndpoint()
    host_endpoint.transfer_endpoint = transfer_endpoint
    host_endpoint.transfer_server_list = transfer_endpoint["DATA"]

    click.echo("Parsing endpoint roles...")
    try:
        res = api.get(f"/endpoint/{epid}/role_list")
    except globus_sdk.TransferAPIError as e:
        raise click.ClickException(e.message)

    host_endpoint.role_list_document = res.data

    if host_endpoint.nodes and len(host_endpoint.nodes) == 1:
        host_endpoint.nodes[0].gridftp_config = gridftp_config

    migration_plan.set_host_endpoint(host_endpoint)

    authorization: Optional[AuthorizeSharing] = None
    if migration_plan.gsi_authz_configuration:
        authorization = AuthorizeSharing(migration_plan.gsi_authz_configuration)
    else:
        click.echo(
            "WARNING: Unable to parse GSI authorization configuration, local "
            "users for guest collections can not be determined",
            err=True,
        )

    storage_gateway = StorageGateway()
    storage_gateway.gridftp_config = gridftp_config
    migration_plan.storage_gateway = storage_gateway

    mapped_collection = MappedCollection()
    mapped_collection.gridftp_config = gridftp_config
    if transfer_endpoint.get("subscription_id"):
        mapped_collection.enable_https = True
    else:
        mapped_collection.enable_https = False
    migration_plan.mapped_collection = mapped_collection

    click.echo("Enumerating shared endpoints...")
    has_next_page = True
    next_page_params: Dict[str, str] = {}
    shared_endpoints = []
    while has_next_page:
        res = api.get(f"/endpoint/{epid}/shared_endpoint_list", params=next_page_params)
        next_token = res.get("next_token")
        has_next_page = next_token is not None
        next_page_params = {"next_token": next_token}
        shared_endpoints += [e["id"] for e in res.get("shared_endpoints", [])]

    sharing_state_dir = gridftp_config.get_string_option("sharing_state_dir")
    if not sharing_state_dir:
        sharing_state_dir = "$HOME/.globus/sharing"

    click.echo("Parsing shared endpoint state...")
    migration_plan.guest_collections = []
    with click.progressbar(
        shared_endpoints,
        width=30,
        show_eta=False,
        item_show_func=lambda x: cast(str, x),
    ) as bar:
        for share in cast(Iterable, bar):
            res = api.get(f"/endpoint/{share}")
            share_transfer_endpoint = res.data

            res = api.get(f"/endpoint_manager/endpoint/{share}/access_list")
            access_list = res.data

            res = api.get(f"/private/endpoint/{share}/anonymous_effective_permissions")
            anonymous_effective_permissions = res.data

            gc = GuestCollection()
            gc.transfer_endpoint = share_transfer_endpoint
            gc.access_list = access_list
            gc.anonymous_effective_permissions = anonymous_effective_permissions

            assert gc.user_cert_chain is not None

            username: Optional[str] = None
            subject: Optional[str] = None
            idx: Optional[int] = None

            if authorization:
                username, subject, idx = authorization.map_and_authorize_sharing(
                    gc.user_cert_chain
                )
                if username is None:
                    click.echo(
                        f"Unable to determine local mapping for guest collection {gc.id}",
                        err=True,
                    )

            credential_data: CredentialDict = {
                "username": username,
                "subject": subject,
                "gsi_authz_callout_index": idx,
                "identity_id": None,
                "identity_username": None,
                "mapping": None,
            }
            gc.credential = Credential(credential_data)

            share_state_file = (
                pathlib.Path(sharing_state_dir)
                / f"share-{anonymous_effective_permissions['sharing_uuid']}"
            )

            gc.share_state_file = share_state_file
            migration_plan.guest_collections.append(gc)

    click.echo("Writing endpoint state to disk...")
    migration_plan.dump()


if __name__ == "__main__":
    create()
