from typing import Dict, TYPE_CHECKING
from typing_extensions import Literal

import click

from ..models import MigrationPlan
from ..util import get_api


if TYPE_CHECKING:
    import globus_sdk
    from ..typing.auth_documents import IdentityDocumentDict

user_cache: "Dict[str, IdentityDocumentDict]" = {}


def lookup_user(
    auth_client: "globus_sdk.AuthClient", username: str
) -> "IdentityDocumentDict":
    global user_cache
    username = username.lower()

    if username not in user_cache:
        res = auth_client.get("/v2/api/identities", params={"usernames": [username]})
        identities = res["identities"]
        if len(identities) != 1:
            raise click.ClickException(f"Unable to resolve {username}")

        if identities[0].get("identity_type"):
            user_cache[username] = identities[0]
    return user_cache[username]


DnNameType = Literal["myproxy", "cilogon", "gridmap"]


@click.command("edit-credentials")
@click.option(
    "--dn-type",
    type=click.Choice(["myproxy", "cilogon", "gridmap"], case_sensitive=False),
    required=True,
    help="Type of DN to process",
)
@click.option(
    "--dn",
    help="Certificate DN. For MyProxy, this is the DN of "
    "the MyProxy CA certificate. For CILogon, this is the prefix used by "
    "CILogon containing the organization. For gridmap, this is the DN of the "
    "user mapping.",
    required=True,
)
@click.option("--domain", metavar="DOMAIN", help="Auth identity domain")
def edit_credentials(dn_type: DnNameType, dn: str, domain: str) -> None:
    migration_plan = MigrationPlan.load()

    # Determine if there's a IdP configured for this domain
    idp = None
    if migration_plan.identity_providers:
        for _idp in migration_plan.identity_providers.values():
            if _idp.domains and domain in _idp.domains:
                idp = _idp
                break
    if not idp:
        if not (
            migration_plan.oidc_server
            and migration_plan.oidc_server.server_name == domain
        ):
            raise click.ClickException(f"No configured idp for {domain}")

    if not migration_plan.guest_collections:
        return
    if (
        not migration_plan.gsi_authz_configuration
        or not migration_plan.gsi_authz_configuration.callouts
    ):
        return

    auth_client = get_api(service="auth")

    if dn_type == "myproxy":
        if (
            migration_plan.gsi_authz_configuration
            and migration_plan.gsi_authz_configuration.callouts
        ):
            for idx, myproxy_config in enumerate(
                migration_plan.gsi_authz_configuration.myproxy_callouts
            ):
                if (
                    myproxy_config
                    and myproxy_config.dn == dn
                    and migration_plan.guest_collections
                ):
                    for gc in migration_plan.guest_collections:
                        if (
                            gc.credential
                            and gc.id
                            and gc.credential.gsi_authz_callout_index == idx
                        ):
                            globus_user = f"{gc.credential.username}@{domain}".lower()
                            if idp:
                                identity = lookup_user(auth_client, globus_user)
                                gc.credential.identity_username = identity["username"]
                                gc.credential.identity_id = identity["id"]
                            else:
                                # OIDC server doesn't yet have ids assigned
                                gc.credential.identity_username = globus_user
                            migration_plan.dump_guest_collection(gc.id)
    elif dn_type == "cilogon":
        if (
            migration_plan.gsi_authz_configuration
            and migration_plan.gsi_authz_configuration.callouts
        ):
            for idx, cilogon_config in enumerate(
                migration_plan.gsi_authz_configuration.cilogon_callouts
            ):
                if cilogon_config.dn == dn:
                    for gc in migration_plan.guest_collections:
                        cred = gc.credential
                        if (
                            not gc.id
                            or not cred
                            or not cred.username
                            or cred.gsi_authz_callout_index is None
                        ):
                            continue
                        if cred.gsi_authz_callout_index == idx:
                            globus_user = f"{cred.username}@{domain}".lower()
                            if idp:
                                identity = lookup_user(auth_client, globus_user)
                                cred.identity_username = identity["username"]
                                cred.identity_id = identity["id"]
                            else:
                                # OIDC server doesn't yet have ids assigned
                                cred.identity_username = globus_user
                            migration_plan.dump_guest_collection(gc.id)
    elif dn_type == "gridmap":
        for idx, gridmap_config in enumerate(
            migration_plan.gsi_authz_configuration.gridmap_configuration
        ):
            if not gridmap_config.mappings:
                continue
            for mapping_dn in gridmap_config.mappings.keys():
                if mapping_dn == dn:
                    for gc in migration_plan.guest_collections:
                        if gc.id is None:
                            continue
                        cred = gc.credential

                        if (
                            not cred
                            or cred.gsi_authz_callout_index is None
                            or cred.username is None
                        ):
                            continue

                        if cred.gsi_authz_callout_index == idx:
                            globus_user = f"{cred.username}@{domain}".lower()
                            if idp:
                                identity = lookup_user(auth_client, globus_user)
                                cred.identity_username = identity["username"]
                                cred.identity_id = identity["id"]
                            else:
                                # OIDC server doesn't yet have ids assigned
                                cred.identity_username = globus_user
                            migration_plan.dump_guest_collection(gc.id)
