import os
import pathlib
from typing import List, Optional, TYPE_CHECKING, Union
from typing_extensions import TypedDict

from ..util import yaml, yaml_object

if TYPE_CHECKING:
    from ..typing.gcs_documents import ExpressionIdentityMappingDict
    from ..typing.auth_documents import IdentityProviderDocumentDict


class AuthIdPDict(TypedDict):
    id: Optional[str]
    domains: Optional[List[str]]
    name: Optional[str]
    short_name: Optional[str]
    alternative_names: Optional[List[str]]
    mapping: Optional["ExpressionIdentityMappingDict"]


@yaml_object(yaml)
class AuthIdP:
    def __init__(self) -> None:
        self.id: Optional[str] = None
        self.mapping: Optional["ExpressionIdentityMappingDict"] = None
        self.domains: Optional[List[str]] = None
        self.name: Optional[str] = None
        self.short_name: Optional[str] = None
        self.alternative_names: Optional[List[str]] = None
        self.default_mapping = True

    @property
    def domains(self) -> Optional[List[str]]:
        return self._domains

    @domains.setter
    def domains(self, value: Optional[List[str]]) -> None:
        self._domains = value

        if value and not self.mapping:
            self.mapping = self.__default_mapping(value)

    def __default_mapping(self, domains: List[str]) -> "ExpressionIdentityMappingDict":
        domains_list = "|".join([d.replace(".", "\\.") for d in domains])

        return {
            "DATA_TYPE": "expression_identity_mapping#1.0.0",
            "source": "{username}",
            "match": f"(.*)@({domains_list})",
            "output": "{0}",
        }

    def add_domain(self, value: str) -> None:
        if self.domains is None:
            self.domains = [value]
        elif self.mapping == self.__default_mapping(self.domains):
            self.domains.append(value)
            self.mapping = self.__default_mapping(self.domains)
        else:
            self.domains.append(value)

    @property
    def identity_provider_document(self) -> Optional["IdentityProviderDocumentDict"]:
        return self._identity_provider_document

    @identity_provider_document.setter
    def identity_provider_document(
        self, value: Optional["IdentityProviderDocumentDict"] = None
    ) -> None:
        self._identity_provider_document = value

        if value:
            if not self.id:
                self.id = value["id"]
            if not self.domains:
                self.domains = value["domains"]
            if not self.name:
                self.name = value["name"]
            if not self.short_name:
                self.short_name = value["short_name"]
            if not self.alternative_names:
                self.alternative_names = value["alternative_names"]

    @staticmethod
    def from_identity_provider_document(
        document: "IdentityProviderDocumentDict",
    ) -> "AuthIdP":
        obj = AuthIdP()
        obj.identity_provider_document = document

        return obj

    def dump(self, root: "Union[str, os.PathLike[str]]") -> None:
        path = pathlib.Path(root) / f"{self.id}.yaml"
        with open(path, "w") as f:
            yaml.dump(self, f)

    @staticmethod
    def load(root: "Union[str, os.PathLike[str]]", id: str) -> "AuthIdP":
        path = pathlib.Path(root) / f"{id}.yaml"
        with open(path, "r") as f:
            obj = yaml.load(f)
            assert isinstance(obj, AuthIdP)

            return obj

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AuthIdP):
            return NotImplemented

        return self is other or self.__getstate__() == other.__getstate__()

    def __getstate__(self) -> AuthIdPDict:
        return {
            "id": self.id,
            "domains": self.domains,
            "name": self.name,
            "short_name": self.short_name,
            "alternative_names": self.alternative_names,
            "mapping": self.mapping,
        }

    def __setstate__(self, state: AuthIdPDict) -> None:
        self.id = state["id"]
        self.mapping = state["mapping"]
        self.domains = state["domains"]
        self.name = state["name"]
        self.short_name = state["short_name"]
        self.alternative_names = state["alternative_names"]
