import os
import pathlib
from typing import cast, List, Optional, Union

from .endpoint_metadata import (
    EndpointMetadata,
    EndpointMetadataDict,
)
from .node import Node
from .role import Role
from ..util import yaml, yaml_object

from ..typing.transfer_documents import (  # noqa: F401
    EndpointDocumentDict,
    EndpointServerDocumentDict,
    NetworkUseName,
    RoleListDocumentDict,
)
from ..typing.gcs_documents import EndpointDict


class HostEndpointDict(EndpointMetadataDict):
    subscription_id: Optional[str]
    network_use: Optional[NetworkUseName]
    max_concurrency: Optional[int]
    preferred_concurrency: Optional[int]
    max_parallelism: Optional[int]
    preferred_parallelism: Optional[int]
    owner: Optional[str]
    allow_udt: Optional[bool]


@yaml_object(yaml)
class HostEndpoint(EndpointMetadata):
    def __init__(self, data: Optional[HostEndpointDict] = None) -> None:
        super().__init__()
        self.owner: Optional[str] = None
        self.subscription_id: Optional[str] = None
        self.roles: Optional[List[Role]] = None
        self.nodes: Optional[List[Node]] = None
        self.transfer_endpoint: Optional["EndpointDocumentDict"] = None
        self.role_list_document: Optional["RoleListDocumentDict"] = None
        self.transfer_server_list: Optional[List["EndpointServerDocumentDict"]] = None
        self.network_use: Optional[NetworkUseName] = None
        self.max_concurrency: Optional[int] = None
        self.preferred_concurrency: Optional[int] = None
        self.max_parallelism: Optional[int] = None
        self.preferred_parallelism: Optional[int] = None
        self.allow_udt: Optional[bool] = None

        if data:
            self.__setstate__(data)

    @property
    def transfer_endpoint(self) -> Optional["EndpointDocumentDict"]:
        return self._transfer_endpoint

    @transfer_endpoint.setter
    def transfer_endpoint(self, value: Optional["EndpointDocumentDict"]) -> None:
        self._transfer_endpoint = value

        if value:
            if self.owner is None:
                self.owner = value["owner_id"]
            if self.subscription_id is None:
                self.subscription_id = value["subscription_id"]
            if self.network_use is None:
                self.network_use = value["network_use"]
            if self.max_concurrency is None:
                self.max_concurrency = value["max_concurrency"]
            if self.preferred_concurrency is None:
                self.preferred_concurrency = value["preferred_concurrency"]
            if self.max_parallelism is None:
                self.max_parallelism = value["max_parallelism"]
            if self.preferred_parallelism is None:
                self.preferred_parallelism = value["preferred_parallelism"]

    @property
    def role_list_document(self) -> Optional["RoleListDocumentDict"]:
        return self._role_list_document

    @role_list_document.setter
    def role_list_document(self, value: Optional["RoleListDocumentDict"]) -> None:
        self._role_list_document = value

        if self._roles is None and value:
            roles = []
            for r in value["DATA"]:
                role = Role()
                role.role_document = r
                roles.append(role)
            self.roles = roles

    @property
    def roles(self) -> Optional[List[Role]]:
        return self._roles

    @roles.setter
    def roles(self, value: List[Role]) -> None:
        self._roles = value

    @property
    def transfer_server_list(self) -> Optional[List["EndpointServerDocumentDict"]]:
        return self._transfer_server_list

    @transfer_server_list.setter
    def transfer_server_list(
        self, value: Optional[List["EndpointServerDocumentDict"]]
    ) -> None:
        self._transfer_server_list = value
        if self.nodes is None and value:
            nodes = []
            for n in value:
                node = Node()
                node.transfer_endpoint_server = n
                nodes.append(node)
            self.nodes = nodes

    @property
    def nodes(self) -> Optional[List[Node]]:
        return self._nodes

    @nodes.setter
    def nodes(self, value: Optional[List[Node]]) -> None:
        self._nodes = value

    def dump(self, root: Union[str, os.PathLike]) -> None:
        path = pathlib.Path(root) / "host_endpoint"
        if not os.path.exists(path):
            os.makedirs(path, 0o755)
        with open(path / "endpoint.yaml", "w") as f:
            yaml.dump(self, f)
        if self.nodes:
            with open(path / "nodes.yaml", "w") as f:
                yaml.dump(self.nodes, f)
        if self.roles:
            with open(path / "roles.yaml", "w") as f:
                yaml.dump(self.roles, f)
        if self.owner:
            with open(path / "owner.yaml", "w") as f:
                yaml.dump(self.owner, f)

    @staticmethod
    def load(root: Union[str, os.PathLike]) -> "HostEndpoint":
        path = pathlib.Path(root) / "host_endpoint"
        ep = path / "endpoint.yaml"

        with open(ep, "r") as f:
            obj = yaml.load(f)
            assert isinstance(obj, HostEndpoint)

        nodes = path / "nodes.yaml"
        if os.path.exists(nodes):
            with open(nodes, "r") as f:
                node_objs = yaml.load(f)
                assert isinstance(node_objs, list)
                for node_obj in node_objs:
                    assert isinstance(node_obj, Node)
                obj.nodes = node_objs

        roles = path / "roles.yaml"
        if os.path.exists(roles):
            with open(roles, "r") as f:
                role_objs = yaml.load(f)
                assert isinstance(role_objs, list)
                for role_obj in role_objs:
                    assert isinstance(role_obj, Role)
                obj.roles = role_objs

        return obj

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, HostEndpoint):
            return NotImplemented
        return self is other or self.__getstate__() == other.__getstate__()

    def __getstate__(self) -> HostEndpointDict:
        metadata_state: EndpointMetadataDict = super().__getstate__()

        return {
            "subscription_id": self.subscription_id,
            "network_use": self.network_use,
            "max_concurrency": self.max_concurrency,
            "preferred_concurrency": self.preferred_concurrency,
            "max_parallelism": self.max_parallelism,
            "preferred_parallelism": self.preferred_parallelism,
            "allow_udt": self.allow_udt,
            "owner": self.owner,
            **metadata_state,  # type: ignore
        }

    def __setstate__(self, state: HostEndpointDict) -> None:  # type: ignore
        super().__setstate__(state)
        self.subscription_id = state["subscription_id"]
        self.network_use = state["network_use"]
        self.max_concurrency = state["max_concurrency"]
        self.preferred_concurrency = state["preferred_concurrency"]
        self.max_parallelism = state["max_parallelism"]
        self.preferred_parallelism = state["preferred_parallelism"]
        self.allow_udt = state["allow_udt"]
        self.owner = state["owner"]

    def to_json(self) -> EndpointDict:
        result = cast(EndpointDict, super().to_json())

        result["DATA_TYPE"] = "endpoint#1.0.0"
        if self.public is not None:
            result["public"] = self.public
        else:
            result["public"] = True
        result["subscription_id"] = self.subscription_id
        if self.network_use is not None:
            result["network_use"] = self.network_use
        if self.network_use == "custom":
            result["max_concurrency"] = self.max_concurrency
            result["preferred_concurrency"] = self.preferred_concurrency
            result["max_parallelism"] = self.max_parallelism
            result["preferred_parallelism"] = self.preferred_parallelism
        if self.allow_udt is not None:
            result["allow_udt"] = self.allow_udt

        return result

    def __str__(self) -> str:
        return f"<Endpoint {self.display_name}>"
