import os
import pathlib
import shutil
from typing import Dict, List, Optional, Union
from typing import Any
from typing_extensions import TypedDict
import uuid

from .auth_idp import AuthIdP
from .globus_oidc_idp import GlobusOIDCIdP
from .gsi_authz_configuration import GsiAuthzConfiguration
from .guest_collection import GuestCollection
from .host_endpoint import HostEndpoint
from .mapped_collection import MappedCollection
from .storage_gateway import StorageGateway

from ..parsers import GCSConfig, GridFTPOptionParser
from ..util import yaml, yaml_object


class MigrationPlanDict(TypedDict):
    environment: str
    host_endpoint: Optional[HostEndpoint]
    gsi_authz_configuration: Optional[GsiAuthzConfiguration]
    storage_gateway: Optional[StorageGateway]
    mapped_collection: Optional[MappedCollection]
    guest_collections: Optional[List[GuestCollection]]
    oidc_server: Optional[GlobusOIDCIdP]


@yaml_object(yaml)
class MigrationPlan:
    def __init__(self, root: Optional[Union[str, os.PathLike]] = None) -> None:
        self._host_endpoint: Optional[HostEndpoint] = None
        self.gcs_config: Optional[GCSConfig] = None
        self.gridftp_config: Optional[GridFTPOptionParser] = None
        self.gsi_authz_configuration: Optional[GsiAuthzConfiguration] = None
        self.storage_gateway: Optional[StorageGateway] = None
        self.mapped_collection: Optional[MappedCollection] = None
        self.guest_collections: Optional[List[GuestCollection]] = None
        self.identity_providers: Optional[Dict[str, AuthIdP]] = None
        self.oidc_server: Optional[GlobusOIDCIdP] = None
        self.environment: Optional[str] = None
        self.root = pathlib.Path(root if root else ".")

    @property
    def host_endpoint(self) -> Optional[HostEndpoint]:
        return self._host_endpoint

    def set_host_endpoint(self, host_endpoint: HostEndpoint) -> None:
        self._host_endpoint = host_endpoint

    @property
    def environment(self) -> Optional[str]:
        if self._environment:
            return self._environment
        return None

    @environment.setter
    def environment(self, value: Optional[str]) -> None:
        self._environment = value

        if value:
            os.environ["GLOBUS_SDK_ENVIRONMENT"] = value
        elif "GLOBUS_SDK_ENVIRONMENT" in os.environ:
            del os.environ["GLOBUS_SDK_ENVIRONMENT"]

    @property
    def gcs_config(self) -> Optional[GCSConfig]:
        return self._gcs_config

    @gcs_config.setter
    def gcs_config(self, value: Optional[GCSConfig]) -> None:
        self._gcs_config = value

        if value and not self.environment:
            self.environment = value.instance

    @property
    def gridftp_config(self) -> Optional[GridFTPOptionParser]:
        return self._gridftp_config

    @gridftp_config.setter
    def gridftp_config(self, value: Optional[GridFTPOptionParser]) -> None:
        self._gridftp_config = value

        if value and not self.gsi_authz_configuration:
            gsi_authz_conf = value.get_string_option("$GSI_AUTHZ_CONF")
            if gsi_authz_conf:
                gridmap = value.get_string_option("$GRIDMAP")
                self.gsi_authz_configuration = GsiAuthzConfiguration(
                    gsi_authz_conf=gsi_authz_conf,
                    gridmap=gridmap,
                    gridftp_config=value,
                )

    def add_guest_collection(self, guest_collection: GuestCollection) -> None:
        if self.guest_collections is None:
            self.guest_collections = []
        self.guest_collections.append(guest_collection)

    def get_guest_collection_by_id(
        self, value: Union[uuid.UUID, str]
    ) -> Optional[GuestCollection]:
        if isinstance(value, uuid.UUID):
            svalue = str(value)
        else:
            svalue = value
        if self.guest_collections:
            for g in self.guest_collections:
                if g.id == svalue:
                    return g
        return None

    def get_identity_provider_for_domain(self, domain: str) -> Optional[AuthIdP]:
        if not self.identity_providers:
            return None

        for idp in self.identity_providers.values():
            if idp.domains and domain in idp.domains:
                return idp
        return None

    def get_identity_provider(self, id: str) -> Optional[AuthIdP]:
        if self.identity_providers:
            return self.identity_providers.get(id)
        else:
            return None

    def add_identity_provider(self, idp: AuthIdP) -> None:
        if not self.identity_providers:
            self.identity_providers = {}
        if not idp.id:
            raise Exception("Misconfigured AuthIdP: no id set")
        self.identity_providers[idp.id] = idp

    def execute(self) -> None:
        print("Create endpoint")
        print("Create storage gateway")
        print("Create mapped collection")
        print("Set managed")
        print("Create endpoint roles")
        print("Create user credentials")
        print("Create guest colletions")
        print("Create guest colletion roles")

    @property
    def migration_plan_dir(self) -> pathlib.Path:
        return pathlib.Path(self.root) / "migration_plan"

    def dump(self) -> None:
        if os.path.exists(self.migration_plan_dir):
            shutil.rmtree(self.migration_plan_dir, ignore_errors=True)
        os.makedirs(self.migration_plan_dir, mode=0o755)

        environment_file = self.migration_plan_dir / "environment"
        with open(environment_file, "w") as f:
            yaml.dump(self.environment, f)

        self.dump_guest_collections()
        self.dump_host_endpoint()
        self.dump_storage_gateway()
        self.dump_mapped_collection()

        self.dump_identity_providers()

        if self.gsi_authz_configuration:
            self.gsi_authz_configuration.dump(self.migration_plan_dir)
        self.dump_oidc_server()

    def dump_guest_collection(self, id: str) -> None:
        guest_collection = self.get_guest_collection_by_id(id)
        if guest_collection:
            gcdir = self.migration_plan_dir / "guest_collections"
            guest_collection.dump(gcdir)

    def dump_guest_collections(self) -> None:
        if self.guest_collections:
            gcdir = self.migration_plan_dir / "guest_collections"
            if not os.path.exists(gcdir):
                os.makedirs(gcdir, mode=0o755)
            for gc in self.guest_collections:
                gc.dump(gcdir)

    def dump_mapped_collection(self) -> None:
        if self.mapped_collection:
            self.mapped_collection.dump(self.migration_plan_dir)

    def dump_host_endpoint(self) -> None:
        if self._host_endpoint:
            self._host_endpoint.dump(self.migration_plan_dir)

    def dump_storage_gateway(self) -> None:
        if self.storage_gateway:
            self.storage_gateway.dump(self.migration_plan_dir)

    def dump_identity_providers(self) -> None:
        if self.identity_providers:
            idps = self.migration_plan_dir / "identity_providers.yaml"
            with open(idps, "w") as f:
                yaml.dump(self.identity_providers, f)

    def dump_oidc_server(self) -> None:
        if self.oidc_server:
            oidc_server = self.migration_plan_dir / "oidc_server.yaml"
            with open(oidc_server, "w") as f:
                yaml.dump(self.oidc_server, f)

    def load_oidc_server(self) -> None:
        oidc_server = self.migration_plan_dir / "oidc_server.yaml"
        if os.path.exists(oidc_server):
            with open(oidc_server, "r") as f:
                self.oidc_server = yaml.load(f)

    @staticmethod
    def load(root: Optional[Union[str, os.PathLike]] = None) -> "MigrationPlan":
        root = pathlib.Path(root if root is not None else ".")

        if not os.path.exists(root):
            raise Exception("No migration plan to load")

        obj = MigrationPlan(root)

        environment_file = obj.migration_plan_dir / "environment.yaml"
        if os.path.exists(environment_file):
            with open(environment_file, "r") as f:
                obj.environment = yaml.load(f)

        obj._host_endpoint = HostEndpoint.load(obj.migration_plan_dir)
        obj.storage_gateway = StorageGateway.load(obj.migration_plan_dir)
        obj.mapped_collection = MappedCollection.load(obj.migration_plan_dir)

        gcdir = obj.migration_plan_dir / "guest_collections"

        guest_collections = []
        if os.path.exists(gcdir):
            for gcid in os.listdir(gcdir):
                gc = GuestCollection.load(gcdir, gcid)
                if gc:
                    guest_collections.append(gc)
            obj.guest_collections = guest_collections

        obj.gsi_authz_configuration = GsiAuthzConfiguration.load(obj.migration_plan_dir)

        obj.load_oidc_server()

        idps = obj.migration_plan_dir / "identity_providers.yaml"
        if os.path.exists(idps):
            with open(idps, "r") as f:
                obj.identity_providers = yaml.load(f)

        return obj

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MigrationPlan):
            return NotImplemented

        return self.__getstate__() == other.__getstate__()

    def __getstate__(self) -> MigrationPlanDict:
        return {
            "host_endpoint": self._host_endpoint,
            "gsi_authz_configuration": self.gsi_authz_configuration,
            "storage_gateway": self.storage_gateway,
            "mapped_collection": self.mapped_collection,
            "guest_collections": self.guest_collections,
            "oidc_server": self.oidc_server,
            "environment": self.environment if self.environment else "default",
        }

    def __setstate__(self, state: MigrationPlanDict) -> Any:
        self._host_endpoint = state["host_endpoint"]
        self.gsi_authz_configuration = state["gsi_authz_configuration"]
        self.storage_gateway = state["storage_gateway"]
        self.mapped_collection = state["mapped_collection"]
        self.guest_collections = state["guest_collections"]
        self.oidc_server = state["oidc_server"]
        self.environment = state["environment"]
