from typing import List, Optional

import pytest

from gcs_migration.models import MigrationPlan, AuthIdP

from gcs_migration.check.migration_plan import (
    MigrationPlanCheck,
    MP01,
    MP02,
    MP03,
    MP04,
    MP05,
)


def test_check_mp01(migration_plan: MigrationPlan) -> None:
    mp01 = MP01()
    checker = MigrationPlanCheck()

    assert not mp01.is_ok(None)
    assert mp01.is_ok(migration_plan)

    res = checker.check(None)
    assert list(filter(lambda x: isinstance(x, MP01), res)) != []

    res = checker.check(migration_plan)
    assert list(filter(lambda x: isinstance(x, MP01), res)) == []


def test_check_mp02(migration_plan: MigrationPlan) -> None:
    mp02 = MP02()
    checker = MigrationPlanCheck()

    migration_plan.storage_gateway = None

    assert mp02.is_ok(None)
    assert not mp02.is_ok(migration_plan)

    res = checker.check(None)
    assert list(filter(lambda x: isinstance(x, MP02), res)) == []

    res = checker.check(migration_plan)
    assert list(filter(lambda x: isinstance(x, MP02), res)) != []


def test_check_mp03(migration_plan: MigrationPlan) -> None:
    mp03 = MP03()
    checker = MigrationPlanCheck()

    migration_plan.mapped_collection = None

    assert mp03.is_ok(None)
    assert not mp03.is_ok(migration_plan)

    res = checker.check(None)
    assert list(filter(lambda x: isinstance(x, MP03), res)) == []

    res = checker.check(migration_plan)
    assert list(filter(lambda x: isinstance(x, MP03), res)) != []


@pytest.mark.parametrize(
    "allowed_domains,auth_domains,oidc_domain,expect",
    [
        ([], [], None, True),
        (["example.org"], ["example.org"], None, True),
        (["example.org"], ["other.org"], None, False),
        (["example.org"], [], "example.org", True),
        (["other.org"], ["example.org"], "example.edu", False),
        (["other.org", "example.org"], ["example.org"], "example.edu", False),
        (["other.org", "example.edu"], ["example.org"], "example.edu", False),
        (
            ["other.org", "example.edu"],
            ["other.org", "example.org"],
            "example.edu",
            True,
        ),
    ],
)
def test_check_mp04(
    migration_plan: MigrationPlan,
    auth_idp: AuthIdP,
    allowed_domains: List[str],
    auth_domains: List[str],
    oidc_domain: Optional[str],
    expect: bool,
) -> None:
    mp04 = MP04()
    checker = MigrationPlanCheck()

    assert migration_plan.storage_gateway is not None
    assert migration_plan.oidc_server is not None

    migration_plan.storage_gateway.allowed_domains = allowed_domains
    migration_plan.identity_providers = None

    if auth_domains:
        auth_idp.domains = auth_domains
        migration_plan.add_identity_provider(auth_idp)

    if oidc_domain:
        migration_plan.oidc_server.server_name = oidc_domain
    else:
        migration_plan.oidc_server = None

    sg = migration_plan.storage_gateway
    assert sg is not None
    sg.allowed_domains = allowed_domains

    assert mp04.is_ok(None) is True
    assert mp04.is_ok(migration_plan) == expect

    res = checker.check(None)
    assert list(filter(lambda x: isinstance(x, MP04), res)) == []

    res = checker.check(migration_plan)
    assert (list(filter(lambda x: isinstance(x, MP04), res)) == []) == expect


@pytest.mark.parametrize(
    "domains,user_pattern,expect",
    [
        ([], "{.username}@example.org", True),
        (["example.org"], "{.username}@example.org", True),
        (["other.org"], "{.username}@example.org", False),
        (["other.org,example.org"], "{.username}@example.org", False),
    ],
)
def test_check_mp05(
    migration_plan: MigrationPlan, domains: List[str], user_pattern: str, expect: bool
) -> None:
    mp05 = MP05()
    checker = MigrationPlanCheck()

    assert migration_plan.storage_gateway is not None
    migration_plan.storage_gateway.allowed_domains = domains

    assert migration_plan.guest_collections is not None
    for gc in migration_plan.guest_collections:
        assert gc.credential is not None
        gc.credential.identity_username = user_pattern.format(gc.credential)

    assert mp05.is_ok(None) is True
    assert mp05.is_ok(migration_plan) == expect

    res = checker.check(None)
    assert list(filter(lambda x: isinstance(x, MP05), res)) == []

    res = checker.check(migration_plan)
    assert (list(filter(lambda x: isinstance(x, MP05), res)) == []) == expect
