from typing import Generator
from unittest import mock

import pytest
import responses

from globus_sdk import AuthClient, TransferClient

from gcs_migration.typing.transfer_documents import (
    AccessDocumentDict,
    AccessListDocumentDict,
    AnonymousEffectivePermissionsDict,
    EndpointDocumentDict,
    EndpointServerDocumentDict,
    RoleDocumentDict,
    RoleListDocumentDict,
    SharedEndpointListDict,
)
from gcs_migration.typing.auth_documents import (
    IdentityDocumentDict,
    IdentityProviderDocumentDict,
)


@pytest.fixture
def mock_responses() -> Generator[None, None, None]:
    responses.start()
    yield
    responses.stop()
    responses.reset()


@pytest.fixture
def mock_tokens(
    mock_responses: Generator[None, None, None]
) -> Generator[None, None, None]:
    with mock.patch(
        "gcs_migration.util.globus_client.__tokens", new_callable=dict
    ) as tokens:
        tokens["auth"] = "AUTH_TOKENS"
        tokens["transfer"] = "TRANSFER_TOKENS"

        yield


@pytest.fixture
def mock_get_transfer_api(
    mock_tokens: None,
    host_endpoint_uuid: str,
    share_endpoint_uuid: str,
    access_acl_document: AccessDocumentDict,
    endpoint_server_document: EndpointServerDocumentDict,
    host_endpoint_document: EndpointDocumentDict,
    shared_endpoint_document: EndpointDocumentDict,
    access_role_document: AccessDocumentDict,
    role_document: RoleDocumentDict,
    role_list_document: RoleListDocumentDict,
    anonymous_effective_permissions_document: AnonymousEffectivePermissionsDict,
    access_list_document: AccessListDocumentDict,
    shared_endpoint_list_document: SharedEndpointListDict,
) -> None:
    tc = TransferClient()

    responses.add(
        "GET",
        f"{tc.base_url}endpoint/{host_endpoint_uuid}",
        json=host_endpoint_document,
    )
    responses.add(
        "GET",
        f"{tc.base_url}endpoint/{share_endpoint_uuid}",
        json=shared_endpoint_document,
    )
    responses.add(
        "GET",
        f"{tc.base_url}endpoint/{host_endpoint_uuid}/role_list",
        json=role_list_document,
    )
    responses.add(
        "GET",
        f"{tc.base_url}endpoint/{host_endpoint_uuid}/shared_endpoint_list",
        json=shared_endpoint_list_document,
    )
    responses.add(
        "GET",
        f"{tc.base_url}endpoint_manager/endpoint/{share_endpoint_uuid}/access_list",
        json=access_list_document,
    )
    responses.add(
        "GET",
        f"{tc.base_url}private/endpoint/{share_endpoint_uuid}/anonymous_effective_permissions",
        json=anonymous_effective_permissions_document,
    )
    return None


@pytest.fixture
def unknown_identity_id() -> str:
    return "3f2c5a46-2b93-413d-a0ce-95bd13d075a0"


@pytest.fixture
def mock_get_auth_api(
    mock_tokens: None,
    identity_provider_document: IdentityProviderDocumentDict,
    identity_document: IdentityDocumentDict,
    unused_idp_identity_document: IdentityDocumentDict,
    email_idp_identity_document: IdentityDocumentDict,
    unknown_identity_id: str,
) -> None:
    ac = AuthClient()

    responses.add(
        "GET",
        f"{ac.base_url}v2/api/identities?usernames={identity_document['username']}",
        json={"identities": [identity_document]},
    )
    responses.add(
        "GET",
        f"{ac.base_url}v2/api/identities?ids={identity_document['id']}",
        json={"identities": [identity_document]},
    )
    responses.add(
        "GET",
        f"{ac.base_url}v2/api/identities?ids={unknown_identity_id}",
        json={"identities": []},
    )
    responses.add(
        "GET",
        f"{ac.base_url}v2/api/identities?ids={unknown_identity_id}",
        json={"identities": []},
    )
    responses.add(
        "GET",
        f"{ac.base_url}v2/api/identities?ids={unused_idp_identity_document['id']}",
        json={"identities": [unused_idp_identity_document]},
    )
    responses.add(
        "GET",
        f"{ac.base_url}v2/api/identities?ids={email_idp_identity_document['id']}",
        json={"identities": [email_idp_identity_document]},
    )
    return None
