import os
import pathlib
import pwd
from typing import Any, Callable, Union

import pytest


@pytest.fixture
def sharing_uuid() -> str:
    return "8d64bdbd-529a-4c4c-8df1-93f117d73480"


@pytest.fixture
def share_path() -> str:
    return "/project"


@pytest.fixture
def sharing_local_user() -> str:
    return "username"


@pytest.fixture(
    params=["~/.globus/sharing", "$HOME/.globus/sharing", "$HOME/sharing-$USER"]
)
def sharing_state_dirs(request: Any) -> str:
    return request.param


@pytest.fixture
def sharing_state_dir() -> str:
    return "$HOME/.globus/sharing"


def __resolve_path_vars(
    path: "Union[str, os.PathLike[str]]",
    user: str,
    home: "Union[str, os.PathLike[str]]",
) -> str:
    if isinstance(path, os.PathLike):
        concrete_path = path.__fspath__()
    else:
        concrete_path = path

    if isinstance(home, os.PathLike):
        concrete_home = home.__fspath__()
    else:
        concrete_home = home

    assert type(concrete_path) == str
    assert type(concrete_home) == str

    resolved_str = concrete_path
    resolved_str = resolved_str.replace("$HOME", concrete_home)
    resolved_str = resolved_str.replace("~", concrete_home)
    resolved_str = resolved_str.replace("$USER", user)
    return resolved_str


ShareStateFileFixtureCreator = Callable[
    [str, str, Union[str, os.PathLike], str, bool, bool], str
]


@pytest.fixture
def home_directory(tmp_path: pathlib.Path) -> pathlib.Path:
    return tmp_path / "homedir"


@pytest.fixture
def share_state_file_creator(
    home_directory: pathlib.Path, monkeypatch: Any
) -> ShareStateFileFixtureCreator:
    def _share_state_file_creator(
        sharing_uuid: str,
        sharing_local_user: str,
        sharing_state_dir: Union[str, os.PathLike],
        share_path: str,
        local_user_exists: bool,
        share_state_file_exists: bool,
    ) -> str:
        def getpwnam(user: str) -> object:
            class pwent:
                pw_name = sharing_local_user
                pw_passwd = "*"
                pw_uid = 42
                pw_gid = 42
                pw_gecos = "Local User"
                pw_dir = str(home_directory)
                pw_shell = "/bin/sh"

            if local_user_exists:
                return pwent()
            else:
                raise KeyError()

        monkeypatch.setattr(pwd, "getpwnam", getpwnam)

        sharing_state_file = pathlib.Path(sharing_state_dir) / f"share-{sharing_uuid}"
        resolved_state_dir = __resolve_path_vars(
            os.path.dirname(sharing_state_file), sharing_local_user, home_directory
        )

        resolved_state_file = __resolve_path_vars(
            sharing_state_file, sharing_local_user, home_directory
        )

        if not os.path.exists(resolved_state_dir):
            os.makedirs(resolved_state_dir, 0o755)

        if share_state_file_exists:
            with open(resolved_state_file, "w") as f:
                print(f'share_path "{share_path}"', file=f)
        return str(sharing_state_file)

    return _share_state_file_creator


@pytest.fixture
def share_state_files(
    share_state_file_creator: ShareStateFileFixtureCreator,
    sharing_uuid: str,
    sharing_local_user: str,
    sharing_state_dirs: str,
    share_path: str,
) -> str:
    return share_state_file_creator(
        sharing_uuid, sharing_local_user, sharing_state_dirs, share_path, True, True
    )


@pytest.fixture
def share_state_file(
    share_state_file_creator: ShareStateFileFixtureCreator,
    sharing_uuid: str,
    sharing_local_user: str,
    sharing_state_dir: str,
    share_path: str,
) -> str:
    return share_state_file_creator(
        sharing_uuid, sharing_local_user, sharing_state_dir, share_path, True, True
    )


@pytest.fixture
def share_state_file_bad_user(
    share_state_file_creator: ShareStateFileFixtureCreator,
    sharing_uuid: str,
    sharing_local_user: str,
    sharing_state_dir: str,
    share_path: str,
) -> str:
    return share_state_file_creator(
        sharing_uuid, sharing_local_user, sharing_state_dir, share_path, False, True,
    )


@pytest.fixture
def share_state_file_bad_state(
    share_state_file_creator: ShareStateFileFixtureCreator,
    sharing_uuid: str,
    sharing_local_user: str,
    sharing_state_dir: str,
    share_path: str,
) -> str:
    return share_state_file_creator(
        sharing_uuid, sharing_local_user, sharing_state_dir, share_path, True, False,
    )
